/*
 * powwow  --  mud client with telnet protocol
 *
 * Initially inspired to the Tintin client by Peter Unold, 
 * Powwow contains no Tintin code.
 * The original program Cancan,  written by Mattias Engdegrd (Yorick)
 * (f91-men@nada.kth.se) 1992-94,
 * was greatly improved upon by Vivriel, Thuzzle and Ilie and then
 * transformed from Cancan into Powwow by Cosmos that heavily worked
 * to make it yet more powerful.
 * AmigaDOS porting attempt by Fror.
 * As usual, all the developers are in debt to countless users 
 * for suggestions and debugging.
 *
 * This program is in the public domain.
 */

#define POWWOW_VERSION "0.9.2 (Yorick, Vivriel, Thuzzle, Ilie, Fror, Cosmos 1992-97)"
#define SAVEFILEVER 4
#define HELPNAME "powwow_help"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <memory.h>
#include <unistd.h>

#ifdef AIX
#  include <sys/select.h>
#endif

#include "powwow.h"
#include "list.h"
#include "cmd.h"
#include "term.h"
#include "edit.h"
#include "xeval.h"

#ifdef RAND
#  define get_random random
#  define init_random srandom
#else
#  define get_random lrand48
#  define init_random srand48
#endif

/*         function declarations                */
void mainlupe();
void exec_delays();
void printmotd();
void get_user_input();
void putbackcursor();
void put_word();
void set_param();
void restore_param();
int search_action();
int printstrlen();
void put_marks();
void wrap_print();
void smart_print();
void intrsig();
void termsig();
void pipesig();
void parse_commands();
void denewbiefy();
void unescape();
void showprompt();
void get_remote_output();
void writetofile();

char *strdup();

extern int errno, sys_nerr;
#ifndef FreeBSD
extern char *sys_errlist[];
#endif
extern int select();
extern int strncasecmp();

/*        GLOBALS                */
char *helpname = HELPNAME;
long stat_var[NUMVAR];

struct {
    long num[NUMPARAM];
    char *str[NUMPARAM];
} def_param;

param_stack paramstk;
vars *var;                   /* pointer to all variables */

long received = 0;           /* amount of data received from remote host */
long sent = 0;               /* amount of data sent to remote host */

char can_suspend = 0;        /* 1 if shell has job control */
static char ready = 1;       /* 0 if powwow is parsing something */
static char confirm = 1;     /* 1 if just tried to quit */
char history_done = 0;       /* 1 if last command contained a #history */   
int internal;             /*  1 if internal echo, -1 if something sent to MUD,
			     -2 if #emulate used */
char echo_ext = 1;           /* 1 if text sent to MUD must be echoed */
char echo_key = 1;           /* 1 if binds must be echoed */
char echo_int = 1;           /* 1 if internal messages are not suppressed */
char compactmode = 0;        /* 1 if to clear prompt between remote messages */
char debug = 0;              /* 1 if to echo every line before executing it */
int also_history_words = 0;  /* 1 if to save also history, 2 for words,
			      * 3 for both */
char hostname[LINELEN], portnumber[LINELEN];
char powwow_dir[LINELEN];          /* default directory path of definition files */
char deffile[LINELEN];             /* name and path of definition file */
char helpfile[LINELEN];            /* name and path of help file */
aliasnode *aliases;                /* head of alias list */
actionnode *actions;               /* head of action list */
marknode *markers;                 /* head of mark list */
int a_nice = 0;                    /* default priority of new actions/marks */
keynode *keydefs;                  /* head of key binding list */
delaynode *delays;                 /* head of delayed commands list */
delaynode *dead_delays;            /* head of dead-delayed commands list */
varnode *named_vars[2];            /* head of named variables list */
int num_named_vars[2];             /* number of named variables defined */
int max_named_vars = 100;          /* initial max number of named variables
				    * (automatically increased when needed) */

vtime now;                         /* current time */
vtime start_time;                  /* time of powwow timer startup */
vtime ref_time;                    /* time corresponding to timer == 0 */

clock_t start_clock, cpu_clock;

connsess conn_list[MAXFDSCAN];       /* connections indexed by socket descriptors */
#define CONN_LIST(n) conn_list[n]

int sockfd = -1;                   /* current socket file descriptor;
				    * -1 means no socket */

int main_conn = -1;		   /* socket file descriptor of main connection;
				    * -1 means no socket */

int count_conn = 0;		   /* number of open connections */

static fd_set fdset;               /* set of descriptors to select() on */

int waiting_for_pid;               /* = pid, when executing #! or editor */

char initstr[LINELEN];             /* initial string to send on connect */
char speedwalk = 0;                /* speedwalk on? */

int linemode;                      /* line mode flags (LM_* in powwow.h) */

/* for line editing */
int cols=80, lines=24;    /* screen size */
int col0;                 /* input line offset (= length of prompt) */
int line0;                /* screen line where the input line starts */
char edbuf[LINELEN];      /* line editing buffer */
int edlen;                /* length of current input line */
int pos = 0;              /* cursor position in line */
char promptstr[LINELEN];  /* entire prompt string (for redrawing) */
int promptlen = 0;        /* length of prompt string */
char promptintercept = 0; /* 1 if last prompt was intercepted by an #action */
char verbatim = 0;        /* 1 = don't expand aliases or process semicolons */
char prefixstr[LINELEN];  /* inserted in the editing buffer each time */
char inserted_next[LINELEN];/* inserted in buffer just once */
char flashback = 0;       /* cursor is on excursion and should be put back */
int excursion;            /* where the excursion is */
char wrap = 1;            /* 1 = word wrap active */
char edattrbeg[CAPLEN];   /* starting input line attributes */
char edattrend[CAPLEN];   /* ending input line attributes */

/* ugly debug macro hack, vt100-like only */
#define NOTE(param) (tty_printf("\0337\033[H\033[K"),tty_printf param,\
		     tty_printf("\0338"),tty_flush())

/* history buffer */
char *hist[MAXHIST];    /* saved history lines */
int curline = 0;        /* current history line */
int pickline = 0;        /* line to pick history from */

/* completion list */
char *words[WORDSKEPT];
int wordindex = 0;

FILE *capfile = (FILE *)NULL;        /* capture file or NULL */
FILE *recordfile = (FILE *)NULL;     /* record file or NULL */

void main(argc, argv)
int argc; char *argv[];
{
    char *p;
    int i;
    
    /* initializations */
    initstr[0] = 0;
    bzero(conn_list, sizeof(conn_list));
    
    gettimeofday(&start_time, NULL);
    now.tv_sec  = ref_time.tv_sec  = start_time.tv_sec;
    now.tv_usec = ref_time.tv_usec = start_time.tv_usec;
    start_clock = cpu_clock = clock();
#ifndef NORANDOM
    init_random((int)now.tv_sec);
#endif
    
    paramstk.curr = 0;      /* stack is empty */
    
    if (p = getenv("POWWOWHELP")) {
        strcpy(helpfile, p);
        if (helpfile[strlen(helpfile) - 1] != '/')
	  strcat(helpfile, "/");
        strcat(helpfile, helpname);
        if (access(helpfile, R_OK) == -1 && !access(helpname, R_OK))
	  strcpy(helpfile, helpname);
    } else
	strcpy(helpfile, helpname);

    if (p = getenv("POWWOWDIR")) {
        strcpy(powwow_dir, p);
        if (powwow_dir[strlen(powwow_dir) - 1] != '/')
	  strcat(powwow_dir, "/");
    } else
	powwow_dir[0] = '\0';

    if (var = (vars *)malloc((NUMVAR+NUMPARAM+max_named_vars)*sizeof(vars))) {
	for (i=0; i<NUMVAR; i++) {
	    var[i].num = &stat_var[i];
	    var[i].str = NULL;
	}
	for (i=NUMVAR; i<NUMVAR+NUMPARAM; i++) {
	    var[i].num = &def_param.num[i];
	    var[i].str = def_param.str[i] = NULL;
	}
    }
    else
      syserr("malloc");
    
    tty_bootsrap();
    
#ifdef MOTDFILE
    printmotd();
#endif
    
    if (argc == 1 || argc == 3) {
	tty_add_initial_binds();
	tty_add_walk_binds();
    } else if (argc == 2 || argc == 4) {
        /*
         * assuming first arg is definition file name
         * If three args, first is definition file name,
         * second and third are hostname and port number
         * (they overwrite the ones in definition file)
         */
        set_deffile(argv[1]);
	
        if (access(deffile,R_OK) == -1 || access(deffile,W_OK) == -1) {
            tty_printf("Creating %s\nHost name  :", deffile);
	    tty_flush();
	    tty_gets(hostname, LINELEN);
	    if (hostname[0] == '\n')
	      hostname[0] = '\0';
	    else
	      strtok(hostname, "\n");
            tty_puts("Port number:");
	    tty_flush();
	    tty_gets(portnumber, LINELEN);
	    if (portnumber[0] == '\n')
	      portnumber[0] = '\0';
	    else
	      strtok(portnumber, "\n");
            tty_add_initial_binds();
	    tty_add_walk_binds();
            if (save_settings() < 0)
	      exit(0);
        } else if (read_settings() < 0)
	  exit(0);
    } 
    if (argc == 3 || argc == 4) {
        /* assume last two args are hostname and port number */
        strcpy(hostname, argv[argc - 2]);
        strcpy(portnumber, argv[argc - 1]);
    }
    
    tty_init();
    signal_init();
    tty_window_change();
    
    tty_puts(tty_clreoscr);
    printver();
    mainlupe();
}

/*
 * set definition file:
 * rules: if powwow_dir is set it is searched first.
 *        if file doesn't exist, it is created there.
 * If a slash appears in the name, the powwow_dir isn't used.
 */
void set_deffile(arg)
char *arg;
{
    if (strchr(arg, '/') == NULL && *powwow_dir) {
	strcpy(deffile, powwow_dir);
	strcat(deffile, arg);
	if ((access(deffile, R_OK) == -1 || access(deffile, W_OK) == -1)
	    && !access(arg,R_OK) && !access(arg, W_OK))
	    strcpy(deffile, arg);
    } else
	strcpy(deffile, arg);
}

/*
 * show current version
 */
void printver()
{
    tty_printf("Powwow version %s\n%s (%s, %s)\n", POWWOW_VERSION, 
#if __STDC__
	       "compiled " __TIME__ " " __DATE__,
#else
	       "", 
#endif
#ifdef VT100
	       "vt100-only",
#else
	       "termcaps",
#endif
#ifdef USETERMIO
	       "termios"
#else
	       "BSD sgtty"
#endif
	       );
}

#ifdef MOTDFILE
/*
 * print the message of the day if present
 */
void printmotd()
{
    char line[LINELEN];
    FILE *f = fopen(MOTDFILE, "r");
    if (f) {
        while (fgets(line, LINELEN, f))
	  tty_puts(line);
	fclose(f);
    }
}
#endif

/*
 * main loop.
 */
void mainlupe()
{
    fd_set readfds;
    int err;
    struct timeval timeoutbuf, *timeout;
    long sleeptime;
    
    FD_ZERO(&fdset);
    FD_SET(tty_input_fd, &fdset);
    
    if (*hostname && *portnumber) {
	open_connection("main", (*initstr ? initstr : NULL), hostname, 
		     atoi(portnumber));
	send_term_size();
    }
    
    tty_putc('\n');
    tty_gotoxy(col0 = 0, lines - 2);
    
    tty_puts("Type #help for help.\n");
    line0 = lines - 1;
    
    confirm = 0;
    
    /*                main lupe             */
    
    for (;;) {
	readfds = fdset;
	
	gettimeofday(&now, NULL); 
	sockfd = main_conn;
	exec_delays();
	
	flush_output_buffer(); /* SPLOOOOSH */
	tty_flush();
	
	sleeptime = 0;
	if (delays) {
	    sleeptime = diff_vtime(&delays->when, &now);
	    if (!sleeptime)
	      sleeptime = 1;    /* if sleeptime is less than 1 millisec,
				 * set to 1 millisec */
	}
	if (flashback && (!sleeptime || sleeptime > FLASHDELAY))
	  sleeptime = FLASHDELAY;
	
	if (sleeptime) {
	    timeoutbuf.tv_sec = sleeptime / mSEC_PER_SEC;
	    timeoutbuf.tv_usec = (sleeptime % mSEC_PER_SEC) * uSEC_PER_mSEC;
	    timeout = &timeoutbuf;
	} else
	  timeout = (struct timeval *)NULL;
	
	ready = 1;
	do {
	    err = select(MAXFDSCAN, &readfds, NULL, NULL, timeout);
	} while (err < 0 && errno == EINTR);
	if (err < 0 && errno != EINTR)
	  syserr("select");
	ready = 0;
	
	if (flashback) putbackcursor();
	
	gettimeofday(&now, NULL);
	
	if (FD_ISSET(tty_input_fd, &readfds))  {
	    err--;
	    sockfd = main_conn;
	    get_user_input();
	}
	
	for (sockfd = 0; err && (sockfd < MAXFDSCAN); sockfd++)
	  if (FD_ISSET(sockfd, &readfds)) {
	      err--;
	      get_remote_output();
	  }
    }
}

/*
 * execute the delayed labels that have expired
 * and place them in the disabled delays list
 */
void exec_delays()
{
    delaynode *dying;
    char buf[LINELEN];
    
    while (delays && cmp_vtime(&delays->when, &now) <= 0) {
	dying = delays;           /* remove delayed command from active list */
	delays = dying->next;     /* and put it in the dead one  */
	
	add_node((defnode *)dying, (defnode **)&dead_delays, rev_time_sort);
	
	/* must be moved before executing delay->command
	 * and command must be copied in a buffer
	 * (can't you imagine why? The command may edit itself...)
	 */
	
	tty_clear_input_line(compactmode);
	if (!compactmode) {
	    tty_putc('\n');
	    col0 = 0;
	}
	
	tty_puts(edattrend);
	if (echo_int)
	  tty_printf("#now [%s]\n", dying->command);
	
	error = 0; internal = 0;
	strcpy(buf, dying->command);
	parse_instruction(buf, 0, 0);
	history_done = 0;
	if (internal >= 0) {
	    showprompt();
	    col0 = printstrlen(promptstr);
	}
	else if (internal != -2)
	  promptlen = promptstr[0] = promptintercept = '\0'; 
	
	tty_draw_input_line(); /* it also prints edattrbeg */
    }
}

/*
 * get output from the socket and process/display it.
 */
void get_remote_output()
{
    char buffer[BUFSIZE + 2];        /* allow for a terminating \0 later */
    char *buf, deleteprompt = 0, kludge = 0;
    int got;
    
    got = read_socket(buffer, BUFSIZE - promptlen);
    if (!got) return;
    
    buffer[got]='\0';  /* Safe, there is space. Do it now not to forget it later */
    received += got;
    
#ifdef DEBUGCODE
    /* debug code to see in detail what strange codes come from the server */
    {
	char c, *t;
	buf = buffer + got;
	tty_putc('{');
	for (t = buffer; t < buf; t++) {
	    if ((c = *t) < ' ' || c > '~')
	      tty_printf("[%d]", c);
	    else tty_putc(c);
	}
	tty_puts("}\n");
    }
#endif
    
    if (!CONN_LIST(sockfd).active)
      return; /* process only active connections */
    
    if (linemode & LM_CHAR) {
	/* char-by-char mode: just display output, no fuss */
	tty_puts(buffer);
	return;
    }
    
    /* line-at-a-time mode: process input in a number of ways */
    
    if (capfile) writetofile(capfile, buffer, got);
    
    /*
     * Some servers like to send NULs. Replace them with spaces, and hope
     * for the best. (I really should cut them out entirely)
     */
    for (buf = buffer; buf < buffer+got; buf++)
      if (!*buf)
        *buf = ' ';
    buf = buffer;
    
    /* 
     * Code to merge lines from host that were splitted
     * into different packets: it is a complete kludge (sigh)
     * and can be used only on one connection at time.
     * We currently do it on main connection.
     */
    
    /*
     * Don't merge if the first part of the line was intercepted
     * by an #action (to avoid intercepting it twice)
     */
    if (sockfd == main_conn && !promptintercept && promptlen && *buf != '\n') {
	/*
	 * put the first part of the splitted line
	 * (currently in the prompt buffer)
	 * on the top of buffer, pretending we have just received it
	 * then clear the prompt buffer
	 */
	bcopy(buf, buf + promptlen, got + 1); /* move also last '\0' */
	memcpy(buf, promptstr, promptlen);
	got += promptlen;
	col0 = promptlen = promptstr[0] = promptintercept = '\0';
	kludge = 1;
    } else if (compactmode && *buf == '\n') {
	/* in compact mode, skip the first \n */
	buf++;
	got--;
	deleteprompt = 1;
	while (got && *buf == '\r') {
	    buf++;     /* some MUDs like to send "\n\r" ... */
	    got--;
	}
    }
    
    
    if (!got)
      return;
    
    tty_puts(edattrend);
    
    if (kludge) {
	/*
	 * try to exit from escape mode:
	 * needed if first part of the line ended with
	 * an unterminated escape sequence
	 tty_putc('\0');
	 */
	/*
	 * now delete the first part of the line,
	 * and parse the whole buffer.
	 */
	tty_gotoxy(col0, line0);
	if (line0 < lines - 1)
	  tty_puts(tty_clreoscr);
	else
	  tty_puts(tty_clreoln);
    } else
      tty_clear_input_line(deleteprompt);
    
    process_remote_input(buf, got);
    
    /*        tty_puts(edattrbeg);   */
    /*  No need to print it, tty_draw_input_line() does it itself */
    
    tty_draw_input_line();
}

/*
 * process input from remote host:
 * detect special sequences, trigger actions, locate prompt, word-wrap,
 * print to tty
 */
void process_remote_input(buf, size)
char *buf; int size;
{
    char onlyprompt = 1, *linestart, *lineend, *end;
    int mpilen = strlen(MPI);
    char buf2[LINELEN];
    
    /* process block one line at a time: */
    
    end = buf + size;
    
    while ((linestart = lineend = buf) < end) {
	/*
	 * process one line at time. stop at "\n" or "\r\n"
	 * also remove extra "\r" before or after "\n"
	 */
	for (; buf < end; buf++) {
	    if (*buf == '\n') {
		lineend = buf;
		*buf = '\0';
		if (buf > linestart && buf[-1] == '\r')
		  buf[-1] = '\0';
		buf++;
		while (buf < end && *buf == '\r')
		  *buf++ = '\0';
		break;
	    }
	}
	if (buf == end && linestart == lineend)
	  lineend = end; /* line continues till end of buffer, no trailing \n */

	if (!strncmp(MPI, linestart, mpilen)) {
	    buf += process_message(linestart + mpilen, end - linestart - mpilen) + mpilen;
	    continue;
	}

#ifdef DEBUGCODE_2
	/* debug code to see in detail what strange codes come from the server */
	{
	    char c;
	    char *t;
	    tty_putc('{');
	    for (t = linestart; t < buf && (c = *t); t++) {
		if (c < ' ' || c > '~')
		  tty_printf("[%d]", c);
		else tty_putc(c);
	    }
	    tty_puts("}\n");
	}
#endif
	
	/*
	 * try to guess where is the prompt... really not much more than
	 * a guess. again, do it only on main connection: we do not want
	 * output from other connections to mess with the prompt
	 * of main connection :)
	 */
	if (sockfd == main_conn && lineend == end && printstrlen(linestart) < cols) {
	    if (onlyprompt && !strchr(linestart, '\b'))
	      strcat(promptstr, linestart);
	    else
	      strcpy(promptstr, linestart);   /* MUME kludge */
	    
	    col0 = printstrlen(promptstr);
	    promptlen = strlen(promptstr);
	    
	    promptintercept = search_action(promptstr); 
	    /* Notice that prompts are still 
	     displayed even if intercepted... */
	    
	    put_marks(buf2, linestart);
	    tty_puts(buf2);
	} else {
	    onlyprompt = 0;
	    if (!search_action(linestart)) {
		if (line0 < lines - 1) line0++;
		if (sockfd != main_conn)        /* sub connection */
		  tty_printf("##%s> ", CONN_LIST(sockfd).id);
		
		smart_print(linestart);
	    }
	    col0 = 0;
	}
    }
}

/*
 * Determine the printed length of a string. This is sometimes less than
 * the string length since it might contain escape sequences. Treated as such
 * are "<esc> [ <non-letters> <letter>", "<esc> <non-[>", "<control-char>".
 * This is not entirely universal but covers the most common cases (i.e. ansi)
 */
int printstrlen(s)
char *s;
{
    int l;
    enum { NORMAL, ESCAPE, BRACKET } state = NORMAL;
    for (l = 0; *s; s++) {
        switch(state) {
	 case NORMAL:
            if (*s == '\033') {
                state = ESCAPE;
            } else if (*s >= ' ') {
                l++;
            }
            break;
	    
	 case ESCAPE:
            state = (*s == '[') ? BRACKET : NORMAL;
            break;
	    
	 case BRACKET:
            if (isalpha(*s))
	      state = NORMAL;
            break;
        }
    }
    return l;
}

/*
 * match action containing &1..&9 and $1..$9 and return actual params in
 * param array - return 1 if matched, 0 if not
 */
int match_action (pat, src)
char *pat, *src;
{
    char mpat[LINELEN], *npat=0, *npat2=0, *nsrc=0, *prm=0, c;
    char *tmp, *line=src;
    int mbeg = 0, mword = 0, p;
    
    for (p = 0; p < 10; p++) {
	*var[p+NUMVAR].str = '\0';
	*var[p+NUMVAR].num = 0;
    }
    
    if (*pat == '^') {
	pat++;
	mbeg = 1;  /* anchor match at line start */
    }
    if (*pat == '&' || *pat == '$')
      mbeg = - mbeg - 1;  /* pattern starts with '&' or '$' */
    
    while (pat && *pat) {
	if (((c=*pat) == '&' || c == '$')) {  /* &x matches a string */
	    /* $x matches a single word */
	    tmp = pat + 1;
	    if (isdigit(*tmp)) {
		p = 0;
		while (isdigit(*tmp) && p < NUMPARAM) {
		    p *= 10;
		    p += *tmp++ - '0';
		}
		if (p <= 0 || p >= NUMPARAM)
		  return 0;
		prm = var[p + NUMVAR].str;
		pat = tmp;
		if (c == '$')
		  mword = 1;
	    }
	}
	
	npat  = first_valid(pat, '&');
	npat2 = first_valid(pat, '$');
	if (npat2 < npat) npat = npat2;
	if (!*npat) npat = 0;
	
	if (npat) {
	    strncpy(mpat, pat, npat-pat);
	    mpat[npat - pat] = 0;
	}
	else
	  strcpy(mpat, pat);
	
	if (*mpat) {
	    nsrc = strstr(src, mpat);
	    if (!nsrc) return 0;
	    if (mbeg > 0) {
		if (nsrc != src) return 0;
		mbeg = 0;  /* reset mbeg to stop further start match */
	    } if (prm) {
		strncpy(prm, src, nsrc-src);
		prm[nsrc - src] = 0;
	    }
	} else if (prm) {
	    /* end of pattern space */
	    strcpy(prm, src);
	}
	
	/* post-processing of param */
	if (prm && mword) {
	    if (mbeg == -1) {
		/* unanchored '$' start, take last word */
		if ((p = my_strrstr(prm, DELIM, strlen(prm) - 1)) != -1) {
		    tmp = prm + p;
		    do
		      *prm++ = *++tmp;
		    while (*tmp);
		}
	    } else if (!pat[1]) {
		/* '$' at end of pattern, take first word */
		strtok(prm, DELIM);
	    } else {
		/* match only if param is single-worded */
		if (my_strstr(prm, DELIM, strlen(prm) - 1) != -1)
		  return 0;
	    }
	    mbeg = mword = 0;  /* reset match flags */
	}
	src = nsrc + strlen(mpat);
	pat = npat;
    }
    strcpy(var[NUMVAR].str, line);
    return 1;
}

/*
 * Search for actions to trig on an input line. The line can't be trashed
 * since we want to print it on the screen later.
 * Return 1 if line matched to some action, 0 otherwise
 */
int search_action(line)
char *line;
{
    actionnode *p;
    params param;
    int ret = 0;
    
    set_param(&param);
    if (error) return 0;
    
    while (*line == '\r') line++;
    for (p = actions; p; p = p->next) {
        if (p->active && match_action(p->pattern, line)) {
            ret = 1;
	    error = 0;
	    internal = -1;
            parse_instruction(p->command, 0, 1);
	    history_done = 0;
            break;
	}
    }
    if (error!=DYN_STACK_UND_ERROR && error!=DYN_STACK_OV_ERROR)
      restore_param();
    return ret;
}

/*
 * match mark containing & and $ and return 1 if matched, 0 if not
 */
int match_mark(pat, src, start, end)
char *pat, *src, **start, **end;
{
    char mpat[LINELEN], *npat=0, *npat2=0, *nsrc=0, *prm=0, *endprm = NULL, *tmp, c;
    int mbeg = 0, mword = 0, p;
    
    *start = *end = 0;
    
    if (*pat == '&' || *pat == '$')
      mbeg = - 1;  /* pattern starts with '&' or '$' */
    
    while (pat && (c = *pat)) {
	if (c == '&' || c == '$') {  /* &x matches a string */
	    /* $x matches a single word */
	    pat++;
	    prm = src;
	    if (c == '$')
	      mword = 1;
	}
	
	npat  = first_valid(pat, '&');
	npat2 = first_valid(pat, '$');
	if (npat2 < npat) npat = npat2;
	if (!*npat) npat = 0;
	
	if (npat) {
	    strncpy(mpat, pat, npat-pat);
	    mpat[npat - pat] = 0;
	} else
	  strcpy(mpat, pat);
	
	if (*mpat) {
	    nsrc = strstr(src, mpat);
	    if (!nsrc) return 0;
	    endprm = nsrc;
	    if (!*start)
	      if (prm)
	      *start = src;
	    else
	      *start = nsrc;
	    *end = nsrc + strlen(mpat);
	} else if (prm)            /* end of pattern space */
	  *end = endprm = prm + strlen(prm);
	else
	  *end = src;
	
	
	/* post-processing of param */
	if (prm && mword) {
	    if (mbeg == -1 && (p = my_strrstr(prm, DELIM, endprm - prm - 1))
		!= -1)
	      /* unanchored '$' start, take last word */
	      *start = prm += p + 1;
	    else if (!*pat)
	      /* '$' at end of pattern, take first word */
	      if ((p = my_strstr(prm, DELIM)) != -1)
	      *end = endprm = prm + p;
	    else
	      /* match only if param is single-worded */
	      for (tmp = prm; tmp < endprm; tmp++)
	      if (strchr(DELIM, *tmp))
	      return 0;
	    mbeg = mword = 0;  /* reset match flags */
	}
	src = nsrc + strlen(mpat);
	pat = npat;
    }
    return 1;
}

/*
 * scan line for marked text and add correct marks
 */
void put_marks(dst, line)
char *dst, *line;
{
    marknode *np, *mfirst;
    char *p, *q, *first, *last;
    char begin[CAPLEN], end[CAPLEN];
    char *endline = line + strlen(line);
    int matchlen;
    
    if (*line) do {
	mfirst = 0;
	first = last = endline;
	for (np = markers; *line && np; np = np->next)
	  if (match_mark(np->pattern, line, &p, &q) && p<first) {
	      
	      /*          if ((p = strstr(line, np->pattern)) && p<first) { */
	      
	      first = p;
	      last = q;
	      mfirst = np;
	  }
	
	if (mfirst) {
	    attr_string(mfirst->attrcode, begin, end);
	    
	    strncpy(dst, line, matchlen = first - line);
	    dst += matchlen;
	    line += matchlen;
	    
	    strcpy(dst, begin);
	    dst += strlen(begin);
	    
	    strncpy(dst, line, matchlen = last - first);
	    
	    dst += matchlen;
	    line += matchlen;
	    
	    strcpy(dst, end);
	    dst += strlen(end);
	}
    } while (mfirst);
    
    strcpy(dst, line);
}

/*
 * write string to tty, wrapping to next line if needed.
 */
void wrap_print(s)
char *s;
{
    char *p, c, follow = 1;
    char buf[LINELEN];
    int l, m;
    enum { NORMAL, ESCAPE, BRACKET } state;
    
    while (*s == '\r') tty_putc(*s++);
    l = printstrlen(s);
    
    while (l >= cols - col0 - 1) {
        p = buf; m = 0; state = NORMAL;
	while (m < cols - col0 - 1 && *s && *s != '\r') {
	    *p++ = c = *s++;
	    switch(state) {
	     case NORMAL:
		if (c == '\033')
		  state = ESCAPE;
		else if (c >= ' ')
		  m++, l--;
		break;
	     case ESCAPE:
		state = (c == '[') ? BRACKET : NORMAL;
		break;
	     case BRACKET:
		if (isalpha(c))
		  state = NORMAL;
		break;
	    }
	}
	
	*p = '\0';
	tty_printf("%s\n", buf);
	col0 = 0;
	if (!*s)
	  follow = 0;
    }
    if (follow)
      tty_printf("%s\n", s);
}

void smart_print(line)
char *line;
{
    char buf[LINELEN];
    
    put_marks(buf, line);
    
    if (wrap)
      wrap_print(buf);
    else
      tty_printf("%s\n", buf);
}

/*
 * write string to file, stripping \r
 */
void writetofile(f, str, len)
FILE *f; char *str; int len;
{
    for (; len--; str++) if (*str != '\r') fputc(*str, f);
}

/*
 * insert a typed character on screen (if it is printable)
 */
void insert_char(c)
char c;
{
    if(edlen < LINELEN - 2 && (c >= ' ' || c < 0)) {
	tty_insert_follow_chars(&c, 1);
	if(ISRPAREN(c)) flashparen(c);
    }
}

/*
 * read terminal input and send to parser.
 * decode keys that send escape sequences
 */
void get_user_input()
{
    int i;
    char c;
    static char typed[CAPLEN];    /* chars typed so far (with partial match) */
    static nchars = 0;                /* number of them */
    
    /* We have 4 possible line modes:
     * line mode, local echo: line editing functions in effect
     * line mode, no echo: sometimes used for passwords, no line editing
     * char mode, no echo: send a character directly, no local processing
     * char mode, local echo: extremely rare, do as above.
     */
    while ((i = tty_read(tty_input_fd, &c, 1)) < 0 && errno == EINTR)
      ;
    if (i != 1)
      syserr("read from tty");
    
    if (linemode & (LM_CHAR | LM_NOECHO)) {
	if (linemode & LM_CHAR) {
	    if (write(sockfd, &c, 1) != 1)
	      syserr("write single char to socket");
	    if (!(linemode & LM_NOECHO)) {
		tty_putc(c);
	    }
	} else {
	    /* sending password (line mode, no echo) */
	    if (c != '\n' && edlen < LINELEN - 2)
		edbuf[edlen++] = c;
	    else {
	        edbuf[edlen] = '\0';
		tty_putc('\n');
		send_to_host(edbuf);
		edlen = 0;
	        typed[nchars = 0] = 0;
	        col0 = promptlen = promptstr[0] = promptintercept = '\0'; 
	    }
	    edbuf[pos = edlen] = '\0';
	}
    } else {
        /* normal mode (echo, line buffered) */
        keynode *p;
	
        confirm = 0;
	
	/* search function key strings for match */
	typed[nchars++] = c;
	for (p = keydefs; p && strncmp(typed, p->sequence, nchars); p = p->next)
	  ;
	if (!p) {
	    insert_char(c);
	    nchars = 0;
	} else if (!p->sequence[nchars]) {
	    p->funct(p->call_data);
	    nchars = 0;
	}
    }
}


/*
 * Flash cursor at parentheses that matches c inserted before current pos
 */
void flashparen(c)
int c;
{
    int lev, i;
    for (i = pos - 1, lev = 0; i >= 0; i--)
      if (ISRPAREN(edbuf[i])) {
	  lev++;
      } else if (ISLPAREN(edbuf[i])) {
	  lev--;
	  if (!lev)
	    if (LPAREN(c) == edbuf[i])
	    break;
	  else
	    i = -1;
      }
    if (i >= 0) {
        tty_gotoxy_opt(CURCOL(pos), CURLINE(pos), CURCOL(i), CURLINE(i));
        flashback = 1;
        excursion = i;
    }
}

/*
 * put cursor back where it belongs
 */
void putbackcursor()
{
    tty_gotoxy_opt(CURCOL(excursion), CURLINE(excursion), CURCOL(pos), CURLINE(pos));
    flashback = 0;
}

/*
 * put string in history at current position
 * (string is assumed to be trashable)
 */
void put_in_history(str)
char *str;
{
    char *p;
    if (hist[curline]) free(hist[curline]);
    hist[curline] = strdup(str);
    if (++curline == MAXHIST) curline = 0;
    
    /* split into words and put into completion list */
    for (p = strtok(str, DELIM); p; p = strtok((char*)NULL, DELIM)) {
        if (strlen(p) >= 3 && p[0] != '#') /* no commands/short words */
	  put_word(p);
    }
}

/*
 * put word in word completion list
 */
void put_word(word)
char *word;
{
    if (words[wordindex]) free(words[wordindex]);
    words[wordindex] = strdup(word);
    if (++wordindex == WORDSKEPT) wordindex = 0;
}

/*
 * match and complete a word referring to the word list
 */
void complete_word()
{
    int i, j;
    
    if (pos) {
	/* find word start */
	for (j = pos; j > 0 && !strchr(DELIM, edbuf[j - 1]); j--);
	if (j == pos) return;
	
        /* scan word list for match */
	for (i = wordindex - 1; i != wordindex; i--) {
	    if (i < 0) i += WORDSKEPT;
	    if (!words[i]) {
		i = wordindex;
		break;
	    }
	    if (!strncasecmp(words[i], edbuf + j, pos - j))
	      break;
	}
	if (i != wordindex) {
	    tty_insert_follow_chars(words[i] + pos - j, strlen(words[i] + pos - j));
	    return;
	}
	
        /* try to complete built-in command */
	
	if (edbuf[j] == '#' && pos - j > 1) {
	    char *p;
	    for (i = 0; (p = commands[i].name); i++) {
		if (!strncmp(p, edbuf + j + 1, pos - j - 1))
		  break;
	    }
	    if (p)
	      tty_insert_follow_chars(p + pos - j - 1, strlen(p + pos - j) + 1);
	}
    }
}

/*
 * match and complete entire lines backwards in history
 */
void complete_line()
{
    int i;
    
    if (edlen) {
	for (i = curline - 1; i != curline; i--) 
	{
	    if (i < 0) i = MAXHIST - 1;
	    if (hist[i] && !strncmp(edbuf, hist[i], edlen)) break;
	}
	if (i != curline) 
	{
	    tty_clear_input_line(0);
	    strcpy(edbuf, hist[i]);
	    pos = edlen = strlen(edbuf);
	    tty_draw_input_line();
	}
    }
}

#ifdef NEWBIE
/*
 * Some people just can't write prose. It's horrible, but here is the
 * fix we have been waiting for.
 */
void denewbiefy()
{
    char *p, *q, *r;
    char buf[LINELEN];
    int n;
    
    strcpy(buf, edbuf);
    p = buf; r = edbuf;
    while (*p) 
    {
	while (*p && !isalnum(*p)) *r++ = *p++;
	q = p;
	while (*p && isalnum(*p)) p++;
	if (*p || *q) 
	{
	    n = p - q;
	    if (!strncasecmp(q, "u", n))
	      strcpy(r, "you"); r += 3;
	    else if (!strncasecmp(q, "ur", n))
	      strcpy(r, "your"); r += 4;
	    else if (!strncasecmp(q, "urs", n))
	      strcpy(r, "yours"); r += 5;
	    else if (!strncasecmp(q, "r", n))
	      strcpy(r, "are"); r += 3;
	    else if (!strncasecmp(q, "4", n))
	      strcpy(r, "for"); r += 3;
	    else if (!strncasecmp(q, "2", n))
	      strcpy(r, "to"); r += 2;
	    else if (!strncasecmp(q, "b", n))
	      strcpy(r, "be"); r += 2;
	    else if (!strncasecmp(q, "b4", n))
	      strcpy(r, "before"); r += 6;
	    else
	      strcpy(r, q); r += n;
	}
    }
    *r = '\0';
    edlen = r - edbuf;
}
#endif

/*
 * put string into input line (will be redrawn next time)
 */
void to_input_line(str)
char *str;
{
    strcpy(inserted_next, str);
}

/* editor utility functions: */

/*
 * send a string to the main connection on the remote host
 */
void send_to_main_host(str)
char *str;
{
    int osockfd = sockfd;  /* backup current socket fd */
    sockfd = main_conn;
    send_to_host(str);
    sockfd = osockfd;
}

/*
 * terminate powwow as cleanly as possible
 */
void exitpowwow()
{
    if (capfile) fclose(capfile);
    if (recordfile) fclose(recordfile);
    tty_puts(edattrend);
    (void)save_settings();
    show_stat();
    tty_quit();
    exit(0);
}

/*
 * set up our signal handlers
 */

void signal_init()
{
    signal(SIGCHLD, childdeathsig);
    signal(SIGINT, intrsig);
    signal(SIGQUIT, intrsig);
    signal(SIGTERM, termsig);
    signal(SIGPIPE, pipesig);
    signal(SIGWINCH, tty_window_change);
    if (signal(SIGTSTP, SIG_IGN) == SIG_DFL) {
	signal(SIGTSTP, suspend_powwow);
	can_suspend = 1;
    }
}

void pipesig()
{
    tty_puts("\n#broken pipe.\n");
    exitpowwow();
}

void intrsig()
{
    if (confirm) {
	tty_puts("\n#interrupted.\n");
	exitpowwow();
    }
    
    tty_puts("\n#interrupted. Press again to quit\n");
    if (ready) {
	showprompt();
	tty_draw_input_line();
    }
    gettimeofday(&now, NULL);
    confirm = 1;
    error = USER_BREAK;
    signal(SIGINT, intrsig);
    signal(SIGQUIT, intrsig);
}

void termsig()
{
    tty_puts("\n#termination signal.\n");
    exitpowwow();
}

/*
 * suspend ourselves
 */
void suspend_powwow()
{
    if (can_suspend) {
        signal(SIGTSTP, SIG_DFL);
        signal(SIGTERM, SIG_IGN);
        signal(SIGINT, SIG_IGN);
        tty_puts(edattrend);
        tty_quit();
	
	if (kill(0, SIGTSTP) == -1)
	  syserr("suspend powwow");
	
        signal_init();
        tty_init();
        tty_window_change();                /* in case size changed meanwhile */
        tty_gotoxy(col0 = 0, line0 = lines - 1);
	/* so we know where the cursor is */
        tty_printf("%s%s%s\n", edattrend, promptstr, edattrbeg);
        tty_draw_input_line();
    } else
      tty_puts("\n#I don't think your shell has job control.\n");
}

/*
 * print system call error message and terminate
 */
void syserr(msg)
char *msg;
{
    if (msg && *msg) {
	fprintf(stderr, "#powwow: system call error:\n %s (%d", msg, errno);
	if (errno > 0 && errno < sys_nerr)
	  fprintf(stderr, ": %s", sys_errlist[errno]);
	fprintf(stderr, ")\n");
    }
    
#ifdef SAVE_ON_SYSERR
    /* Try to do an emergency save. This can wreak havoc
     * if called from the wrong place, like
     * read_settings() or save_settings(),
     * thus is executed only if you add -DSAVE_ON_SYSERR
     * to CF flags in make_it
     */
    (void)save_settings();
#endif
    
    tty_quit();
    exit(1);
}

void set_param(altparam)
params *altparam;
{
    int i;
    
    if (paramstk.curr < MAX_STACK) {
	paramstk.p[paramstk.curr++] = (params *)var[NUMVAR].num;
        /* it works, because var[NUMVAR].num points to param.num[0],
         * which is the first field of param.
         */
	
	for (i=0; i<NUMPARAM; i++) {
	    *(var[i+NUMVAR].num = &altparam->num[i]) = 0;
	    *(var[i+NUMVAR].str = altparam->str[i]) = 0;
	}
    }
    else {
	print_error(error=DYN_STACK_OV_ERROR);
	paramstk.curr = 0;    /* reset stack to empty */
	for (i=0; i<NUMPARAM; i++) {
	    var[i+NUMVAR].num = &def_param.num[i];
	    var[i+NUMVAR].str = def_param.str[i];
	}
    }
}

void restore_param()
{
    int i;
    params *v;
    
    if (paramstk.curr > 0) {
	v = paramstk.p[--paramstk.curr];
	
	for (i=0; i<NUMPARAM; i++) {
	    var[i+NUMVAR].num = &v->num[i];
	    var[i+NUMVAR].str = v->str[i];
	}
    }
    else {
	print_error(error=DYN_STACK_UND_ERROR);
	paramstk.curr = 0;    /* reset stack to empty */
	for (i=0; i<NUMPARAM; i++) {
	    var[i+NUMVAR].num = &def_param.num[i];
	    var[i+NUMVAR].str = def_param.str[i];
	}
    }
}

/*
 * parse input from user: calls parse_instruction for each instruction
 * in cmd_line.
 * silent = 1 if the line should not be echoed, 0 otherwise.
 */
void parse_user_input(line, silent)
char *line; char silent;
{
    do {
	line = parse_instruction(line, silent, 0);
    } while (!error && line && *line);
}

/*
 * execute walk if word is valid [speed]walk sequence -
 * return 1 if walked, 0 if not
 */
int parse_walk (word, silent, maponly)
char *word, silent, maponly;
{
    long strtol();
    char buf[16];
    int n = strlen(word);
    int is_main = (sockfd == main_conn);
    
    if (!is_main && !maponly && !speedwalk)
      return 0;
    if (!n || (n > 1 && !speedwalk && !maponly) ||
	!strchr("neswud", word[n - 1]) ||
	strspn(word, "neswud0123456789") != n)
      return 0;
    
    if (maponly) silent = 1;
    else if (speedwalk) internal = -1;
    buf[1] = '\0';
    while (*word) {
        if (!silent) tty_putc('[');
        if (isdigit(*word)) {
            n = strtol(word, &word, 10);
	    if (!silent)
	      tty_printf("%d", n);
	} else
	  n = 1;
        if (!silent)
	  tty_putc(*word);
        while (n--) {
            *buf = *word;
            if (!maponly) {
	        send_to_host(buf);
	    }
            if (is_main || maponly)
	      add_dir_to_map(*word);
        }
	if (!silent)
	  tty_puts("] ");
        word++;
    }
    if (!silent)
      tty_putc('\n');
    return !maponly;
}

/*
 * split str into words separated by DELIM, and place in
 * var[1+NUMVAR].str ... var[9+NUMVAR].str -
 * the whole str is put in var[0+NUMVAR].str
 */
char *split_words (str)
char *str;
{
    int i;
    char *prm;
    
    strcpy(var[NUMVAR].str, str);
    for (i=NUMVAR + 1; i < NUMPARAM+NUMVAR; i++) {
	*var[i].num = 0;
	prm = var[i].str;
	while (*str && strchr(DELIM, *str))
	  str++;
	while (*str && !strchr(DELIM, *str))
	  *prm++ = *str++;
	*prm = 0;
    }
    return str;
}

/*
 * copy first word of src into dst, and return pointer to second word of src
 */
char *split_first_word(dst, src)
char *dst, *src;
{
    char *tmp, *delim = DELIM;
    int what, got = 0;
    
    tmp = src = skipspace(src);
    
    if (!*src) {
	*dst='\0';
	return src;
    }
    
    do
      for  (what = 0; delim[what] && !got; what++)
      if (*tmp == delim[what])
      got = 1;
    while (!got && *++tmp);
    
    strncpy(dst, src, tmp-src);
    dst[tmp-src] = '\0';
    
    if (*tmp && *tmp != CMDSEP) tmp++;
    return tmp;
}

/*
 * Parse and exec the first instruction in 'line', and return pointer to the
 * second instruction in 'line' (if any).
 */
char *parse_instruction(line, silent, subs)
char *line; char silent, subs;
{
    aliasnode *np;
    char buf[LINELEN], substbuf[LINELEN];
    char *arg, *end, *ret;
    char last_is_sep = 1;
    int osockfd = -1;
    
    if (error) return NULL;
    
    ret = get_next_instr(line);
    
    if (!ret || ret==line)  /* RETURN alone must be sent using a ';' */
      return ret;
    
    if (ret[-1] == CMDSEP)
      *--ret = '\0';
    else if (*ret) {       /* should NEVER happen... */
	tty_printf("#PANIC! Tried to execute: %s\n", line);
	error=SYNTAX_ERROR;
	return NULL;
    }
    else
      last_is_sep = 0;
    
    
    if (subs)
      subst_param(substbuf, line);
    else
      strcpy(substbuf, line);
    
    if (debug) tty_printf("#parsing: %s\n", substbuf); /* For run-time debugging */
    
    if (last_is_sep)
      *ret++ = CMDSEP;
    
    arg = skipspace(line = substbuf);
    
    if (arg[0] == '#' && arg[1] == '#') { /* send to other connection */
	line = split_first_word(buf, arg + 2);
	osockfd = sockfd;
	if ((sockfd = find_connection(buf))<0) {
	    tty_printf ("#no connection named '%s'\n", buf);
	    sockfd = osockfd;
	    return NULL;
	}
	arg = skipspace(line);
	if (!*arg) { /* set it as main connection */
	    main_conn = sockfd;
	    return ret; /* stop parsing, otherwise a newline would be sent to sockfd */
	}
    }
    
    if (*arg == '{') {    /* instruction contains a block */
	end = first_regular(line = arg + 1, '}');
	
	if (*end) {
	    *end = '\0';
	    parse_user_input(line, silent);
	    *end = '}';
	}
	else
	  print_error(error=MISSING_PAREN_ERROR);
    }
    else if (*arg == '#') {
	if (*(end = skipspace(arg + 1)) == '(') {    /* execute #() */
	    end++;
	    (void)xeval(NULL, &end, PRINT_NOTHING);
	}
	else
	  parse_commands(buf, split_first_word(buf, arg + 1));
    } else {
	int oneword;
	/* initial spaces are NOT skipped this time */
	
	arg = split_first_word(buf, line); /* buf contains the first word,
					    arg points to arguments */
	if (!*arg) oneword = 1;
	else oneword = 0;
	
	if (np = *lookup_alias(buf)) {
	    params param;
	    set_param(&param);
	    if (error) return NULL;
	    
	    split_words(arg);  /* split argument into words
                                and place them in $0 ... $9 */
	    parse_instruction(np->subst, 0, 1);
	    
	    if (error!=DYN_STACK_UND_ERROR && error!=DYN_STACK_OV_ERROR)
	      restore_param();
	    
	} else if (!oneword || !parse_walk(buf, silent, 0)) {
	    /* it is ok, parse_walk accepts only one word */
	    
	    unescape(line);
	    if (!silent && echo_ext) tty_printf("[%s]\n", line);
	    col0 = promptlen = promptstr[0] = promptintercept = '\0'; 
	    send_to_host(line);
	}
    }
    if (osockfd != -1)
      sockfd = osockfd;
    
    return ret;
}

/*
 * remove escapes (backslashes) from a string
 */
void unescape(s)
char *s;
{
    char *p;
    p = s;
    while (*s) {
	if (*s == '\\') {
	    s++;
	    while (*s == '\\')
	      *p++ = *s++;
	    if (*s == NIL_CHAR) {
		*p++ = '\\';
		s++;
		continue;
	    }
	}
	if (*s != '\\')
	  *p++ = *s++;
    }
    *p = '\0';
}

/*
 * parse powwow's own commands
 */
void parse_commands(command, arg)
char *command, *arg;
{
    int i, j;
    
    /* We ALLOW special commands also on subsidiary connections ! */
    
    /* assume output will be enough to make input line = last screen line */
    line0 = lines - 1;
    if (isdigit(*command) && (i = atoi(command)))
      if (i > 0)
      while (i--) (void)parse_instruction(arg, 1, 0);
    else
      tty_puts("#bogus repeat count\n");
    else {
	j = strlen(command);
	for (i = 0; commands[i].name; i++)
	  if (!strncmp(command, commands[i].name, j)) {
	      if (commands[i].funct) {
		  tty_puts(edattrend);
		  (*commands[i].funct)(arg);
		  return;
	      }
	  }
	tty_printf("#unknown powwow command '%s'\n", command);
    }
}

/*
 * redisplay the prompt
 */
void showprompt()
{
    char buf[LINELEN];
    put_marks(buf, promptstr);
    tty_printf("%s%s%s", edattrend, buf, edattrbeg);
}

/*
 * cancel an editing session; does not free anything
 * (the child death signal handler will remove the session from the list)
 */
void cancel_edit(sp)
editsess *sp;
{
    char buf[LINELEN];
    char keystr[16];
    
    if (kill(sp->pid, SIGKILL) < 0)        /* Editicide */
      syserr("kill editor child");
    tty_printf("#killed %s (%d)\n", sp->descr, sp->key);
    sprintf(keystr, "C%d\n", sp->key);
    sprintf(buf, "%sE%d\n%s", MPI, (int) strlen(keystr), keystr);
    send_to_main_host(buf);
    sp->key = -1;        /* mark the child as cancelled */
}

/*
 * substitute $0..$9 and @0..@9 in a string 
 * (unless $ or @ is escaped with backslash)
 */
void subst_param(dst, src)
char *src, *dst;
{
    int i, subst;
    char *tmp, c;
    
    while (*src)  {
	subst = 0;
	if (*src == '$' || *src == '@') {
	    c = *src;
	    tmp = src + 1;
	    if (isdigit(*tmp)) {
		i = 0;
		while (isdigit(*tmp) && i < NUMPARAM) {
		    i *= 10;
		    i += *tmp++ - '0';
		}
		if (i < NUMPARAM) {
		    src = tmp;
		    subst = 1;
		    if (c == '$') {
			strcpy(dst, var[i+NUMVAR].str);
			if (*dst) while (*++dst) ;
		    } else {
			sprintf(dst, "%ld", *var[i+NUMVAR].num);
			while (*++dst) ;
		    }
		}
	    }
	} 
	if (!subst) {
	    if (*src == '\\') {
		src++;
		while (*src == '\\')
		  *dst++ = *src++;
		
		if (*src)
		  *dst++ = *src++;
	    }
	    else if (*src)
	      *dst++ = *src++;
	}
    }
    *dst = '\0';
}

/*
 * return pointer to next non-blank char
 */
char *skipspace(p)
char *p;
{
    while (*p == ' ') p++;
    return p;
}

/*
 * read definitions from file
 * return > 0 if success, < 0 if fail.
 * NEVER call syserr() from here or powwow will
 * try to save the definition file even if it got
 * a broken or empty one.
 */
int read_settings()
{
    char buf[LINELEN];
    FILE *f;
    char *p, *cmd, done, old_echo_int = echo_int, old_nice = a_nice;
    int failed = 1, n;
    
    if (!*deffile) {
	tty_puts("#warning: no save-file defined!\n");
	return -1;
    }
    
    f = fopen(deffile, "r");
    if (!f) {
	PRINT_INT ("#cannot find save-file: %s\n", deffile);
	return -1;
    }
    
    n = echo_int = a_nice = 0;
    
    while (aliases)       delete_aliasnode(&aliases);
    while (actions)       delete_actionnode(&actions);
    while (markers)       delete_marknode(&markers);
    while (keydefs)       delete_keynode(&keydefs);
    while (named_vars[0]) delete_varnode(&named_vars[0], (char)0);
    while (named_vars[1]) delete_varnode(&named_vars[1], (char)1);
    
    for (n = 0; n < NUMVAR; n++) {
	*var[n].num = 0;
	if (p = var[n].str) {
	    free(p);
	    var[n].str = NULL;
	}
    }
    
    while (fgets(buf, LINELEN, f)) {
	cmd = strtok(buf, "\n");        /* strip lf */
	if (!cmd)
	  continue;
        cmd = skipspace(cmd);
	if (!*cmd)
	  continue;
        error = 0;
	done = 0;
        if (*(p = first_regular(cmd, ' '))) {
	    done = 1;
            *p++ = '\0';
            if (!strcmp(cmd, "#savefile-version")) {
                n = atoi(p);
                if (n > SAVEFILEVER)
		  tty_puts("#warning: this powwow version is too old!\n");
                else if (n < SAVEFILEVER)
		  tty_puts("#warning: config file is from an older version\n");
            } else {
	        *--p = ' ';
		done = 0;
	    }
	}
	if (!done)
	  parse_user_input(cmd, 1);
    }
    
    if (ferror(f) && !feof(f)) {
        tty_printf("#error: can't read file '%s': %s\n", deffile, strerror(errno));
        n = 0; failed = -1;
    }
    
    fclose(f);
    if (n < 1) {
        tty_puts("#default keypad settings loaded\n");
        tty_add_walk_binds();
    }
    if (n < 2) {
        tty_puts("#default editing keys settings loaded\n");
        tty_add_initial_binds();
    }
    echo_int = old_echo_int;
    a_nice = old_nice;
    return failed;
}

static char tmpname[LINELEN];

void fail_msg()
{
    tty_printf("#error: can't write to temporary file '%s': %s\n", tmpname, strerror(errno));
}

/*
 * save settings in definition file
 * return > 0 if success, < 0 if fail.
 * NEVER call syserr() from here or powwow will
 * enter an infinite loop!
 */
int save_settings()
{
    FILE *f;
    int l;
    aliasnode *alp;
    actionnode *acp;
    marknode *mp;
    keynode *kp;
    varnode *vp;
    
    char buf[LINELEN], *buf2;
    int i, flag, failed = 1;
    
    if (!*deffile) {
	PRINT_INT("#warning: no save-file defined!\n");
	return -1;
    }
    
    /*
     * Create a temporary file in the same directory as deffile,
     * and write settings there
     */
    strcpy(tmpname, deffile);
    l = strlen(tmpname) - 1;
    while (l && tmpname[l] != '/')
      l--;
    if (l)
      l++;
    
    sprintf(tmpname + l, "tmpsav%d%d", getpid(), rand() >> 8);
    if (!(f = fopen(tmpname, "w"))) {
        fail_msg();
        return -1;
    }
    
    failed = fprintf(f, "#savefile-version %d\n", SAVEFILEVER);
    if (failed > 0)
      failed = fprintf(f, "#host %s %s\n", hostname, portnumber);
    /* ************************
     if (failed > 0)
     failed = fprintf(f, "#option %s%s\n",
		      also_history_words & 1 ? "history " : "",
		      also_history_words & 2 ? "words" :
		      also_history_words ? "" : "none");
     ************************ */
    
    /* main host */
    if (failed > 0 && *initstr)
      failed = fprintf(f, "#init =%s\n", initstr);
    
    for (alp = aliases; alp && failed > 0; alp = alp->next) {
	escape_specials(alp->name, buf2=buf);
	failed = fprintf(f, "#alias %s=%s\n", buf, alp->subst);
    }
    
    for (acp = actions; acp && failed > 0; acp = acp->next) {
        escape_specials(acp->label, buf2=buf);
        while (*buf2) buf2++;
        *buf2++ = ' ';
        escape_specials(acp->pattern, buf2);
        while (*buf2) buf2++;
	failed = fprintf(f, "#action >%c%s=%s\n", acp->active ? '+' : '-',
			 buf, acp->command);
    }
    
    for (mp = markers; mp && failed > 0; mp = mp->next) {
        escape_specials(mp->pattern, buf);
	failed = fprintf(f, "#mark %s=%s\n", buf, attr_name(mp->attrcode));
    }
    /* save value of global variables */
    
    for (flag = 0, i=0; i<NUMVAR && failed > 0; i++) {
        if (var[i].num && *var[i].num) {    /* first check was missing!!! */
	    failed = fprintf(f, "%s@%d = %ld", flag ? ", " : "#(",
			     i-NUMVAR, *var[i].num);
            flag = 1;
        }
    }
    if (failed > 0 && flag) failed = fprintf(f, ")\n");
    
    for (i=0; i<NUMVAR && failed > 0; i++) {
        if (var[i].str && *var[i].str) {    /* first check was missing!!! */
	    escape_specials(var[i].str, buf);
	    failed = fprintf(f, "#($%d = \"%s\")\n", i-NUMVAR, buf);
	}
    }
    
    for (flag = 0, vp = named_vars[0]; vp && failed > 0; vp = vp->next) {
        if (vp->num) {
	    failed = fprintf(f, "%s@%s = %ld", flag ? ", " : "#(",
			     vp->name, vp->num);
            flag = 1;
	}
    }
    if (failed > 0 && flag) failed = fprintf(f, ")\n");
    
    for (vp = named_vars[1]; vp && failed > 0; vp = vp->next) {
        if (*vp->str) {
	    escape_specials(vp->str, buf);
	    failed = fprintf(f, "#($%s = \"%s\")\n", vp->name, buf);
	}
    }
    
    /* *********************
     if (also_history_words & 1) {
	 for (l = curline - 1; l != curline && failed > 0; l--) {
	     if (l < 0) l += MAXHIST;
	     if (!hist[l])
	       break;
	     else {
		 escape_specials(hist[i], buf);
		 failed = fprintf(f, "#put %s\n", buf);
	     }
	 }
     }
     if (also_history_words & 2) {
	 for (flag = 0, l = wordindex - 1; l != wordindex && failed > 0; l--) {
	     if (l < 0) l += WORDSKEPT;
	     if (!words[l])
	       break;
	     else {
		 escape_specials(words[l], buf);
		 failed = fprintf(f, "%s %s", flag ? "" : "#add", buf);
	     }
	 }
	 if (failed > 0 && flag)
	   failed = fprintf(f, "\n");
     }
     ********************* */
    
    for (kp = keydefs; kp && failed > 0; kp = kp->next) {
	escape_specials(seq_name(kp->sequence), buf);
	failed = fprintf(f, "#bind %s %s=%s\n", kp->name, buf,
			 kp->funct==key_run_command ? kp->call_data :
			 internal_functions[lookup_edit_function(kp->funct)].name);
    }
    
    fclose(f);
    
    if (failed <= 0) {
	failed = -1;
	fail_msg();
    }   
    else {
	failed = rename(tmpname, deffile);
	if (failed == -1) {
	    PRINT_INT ("#error: can't move temporary file '%s' to '%s': %s\n",
		       tmpname, deffile, strerror(errno));
	} else
	  failed = 1;
    }
    
    return failed;
}

#ifdef NOSTRDUP
/*
 * some old libraries don't have strdup(), so...share and enjoy. 
 */
char *strdup(s)
char *s;
{
    char *p = (char*)malloc(strlen(s) + 1);
    if (p) strcpy(p, s);
    return p;
}
#endif


/*
 * Below are multiple-connection support functions:
 */

/*
 * return connection's index or socket descriptor given id,
 * or -1 if null or invalid id is given
 */
int find_connection (id)
char *id;
{
    int i, n;
    
    for (i=0, n=count_conn; n; i++)
      if (CONN_LIST(i).id) 
    {
        n--;
        if (!strcmp(CONN_LIST(i).id, id))
	  return i;
    }
    return -1;
}



void show_host()
{
    if (!internal) internal = 1;
    if (*hostname)
      tty_printf("#host: %s %s\n", hostname, portnumber);
    else
      tty_puts("#host not defined.\n");
}


/*
 * show list of open connections
 */
void show_connections()
{
    int i, n = count_conn;
    
    PRINT_INT ("#%s connection%s opened%c\n", n ? "The following" : "No", 
	       n==1 ? " is" : "s are", n ? ':' : '.');
    
    
    for (i=0, n=count_conn; n && i < MAXFDSCAN; i++)
      if (CONN_LIST(i).id) {
	  n--;
	  tty_printf("%s ##%s   \t%sactive \t(%s %d)\n",
		     i == main_conn ? "(default)" : "         ",
		     CONN_LIST(i).id, CONN_LIST(i).active ? "   " : "non",
		     CONN_LIST(i).host, CONN_LIST(i).port);
      }
}

/*
 * open another connection
 */
void open_connection (id, initstr, host, port)
char *id;        /* connection id */
char *initstr;        /* optional init string */
char *host; int port;  /* address of remote host */
{
    int newsockfd;
    
    if (find_connection(id)>=0) {
	PRINT_INT ("#connection '%s' already open.\n", id);
	return;
    }
    
    /* dial the number by moving the right index in small circles */
    if ((newsockfd = doconnect(host, port)) < 0) {
        PRINT_INT("#unable to connect to host. Giving up.\n");
	return;
    }
    if (echo_int && count_conn) {
	PRINT_INT("#default connection is now %s\n", id);
    }
    
    FD_SET(newsockfd, &fdset);         /* add socket to select() set */
    count_conn++;
    main_conn = sockfd = newsockfd;             /* ALWAYS set as main connection */
    CONN_LIST(newsockfd).active = 1;
    CONN_LIST(newsockfd).id = strdup(id);
    CONN_LIST(newsockfd).host = strdup(host);
    CONN_LIST(newsockfd).port = port;
    
    gettimeofday(&now, NULL); /* an #at or #in may be inside initstr */
    
    if (initstr)
      parse_instruction(initstr, 0, 0);
}

/*
 * close a connection
 */
void close_connection (id)
char *id;
{
    int i, sfd;
    
    if (id) {  /* #disconnect cmd */
	if ((sfd = find_connection(id))<0) {
	    PRINT_INT("#no such connection: %s\n", id);
	    return;
	}
	close(sfd);
    } else
      sfd = sockfd;  /* connection closed by remote host */
    
    PRINT_INT("#connection on %s closed.\n", CONN_LIST(sfd).id);
    
    if (sfd == main_conn) { /* main connection closed */
	if (count_conn == 1) { /* was last connection */
	    tty_puts("#no connections left. Type #quit to quit.\n");
	    sockfd = main_conn = -1;
	} else {               /* must find another connectiion and promote it to main */
	    for (i = 0; i < MAXFDSCAN; i++) {
		if (i == sfd)
		  continue;
		if (CONN_LIST(i).id) {
		    tty_printf("#default connection is now %s\n", CONN_LIST(i).id);
		    main_conn = i;
		    if (sockfd == sfd)
		      sockfd = -1;
		    break;
		}
	    }
	    if (sfd == main_conn) {
		tty_printf("#internal error in close_connection()\nGiving up.\n");
		syserr(NULL);
	    }
	}
    }
    
    if (sockfd == sfd)
      sockfd = -1; /* no further I/O allowed on sfd, as we just closed it */
    
    FD_CLR(sfd, &fdset);
    count_conn--;
    CONN_LIST(sfd).active = 0;
    free(CONN_LIST(sfd).id);
    CONN_LIST(sfd).id = 0;
}

/*
 * toggle output display from another connection
 */
void snoop_connection (id)
char *id;
{
    int sfd;
    
    sfd = find_connection(id);
    if (sfd>=0) {
        CONN_LIST(sfd).active ^= 1;
	if (echo_int) {
	    PRINT_INT("#connection %s is now %sactive.\n",
		      CONN_LIST(sfd).id, CONN_LIST(sfd).active ? "" : "non");
	}
    } else {
        PRINT_INT("#no such connection: %s\n", id);
    }
}
